# Copyright (c) 2007 Canonical
#
# Written by Jamshed Kakar <jkakar@kakar.ca>
#
# This file is part of AutoPPA.
#
# AutoPPA is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# AutoPPA is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from autoppa.lib.configobj import ConfigObj


class Configuration(object):
    """Get configuration details for named build targets.

    The configuration file, generally stored at C{~/.autoppa.conf},
    contains details used during the build process.  The file contains
    a stanza for each build target to build with PPA.  Configuration
    values stored in this file are expected to change infrequently.
    The following case-sensitive configuration keys must be present
    for each build target stanza in the configuration file:

      * email: This is the e-mail address in 'First Name
            <first.name@example.com>' format.
      * branch: The path of the source branch used to generate sources
            from.
      * repository: The path of the Bazaar repository to generate the
            working tree in.  This is the branch that is created
            during the build process.  It is eventually merged back
            into C{branch} unless merging is suppressed.
      * ppa: The name of the PPA to upload sources to as defined in
            C{~/.dput.cf}.
      * releases: A space-separated list of Ubuntu releases to build
            packages for.

    The following cases-sensitive configuration key can optionally be
    provided:

      * options: A space-separated list of options.  Currently the
            only possible option is C{no-version-replacement}.

    An example configuration file looks like::

      [landscape-client]
      email = Jamshed Kakar <jamshed.kakar@canonical.com>
      branch = /home/jkakar/src/canonical/landscape-client/production
      repository = /home/jkakar/src/canonical/landscape-client
      ppa = landscape-client-ppa
      releases = dapper edgy feisty gutsy

      [smart]
      email = Jamshed Kakar <jamshed.kakar@canonical.com>
      branch = /home/jkakar/src/canonical/smart/trunk
      repository = /home/jkakar/src/canonical/smart
      ppa = landscape-client-ppa
      releases = dapper edgy feisty gutsy
    """

    def __init__(self, configuration):
        self._configuration = configuration

    @classmethod
    def from_file(cls, filename):
        """Create a C{Configuration} instance initialized from C{filename}."""
        return cls(ConfigObj(filename))

    def get_build_target(self, build_target):
        """Get a C{dict}-like object with data related to C{build_target}."""
        try:
            return self._configuration[build_target]
        except KeyError:
            pass
        return None
