# Copyright (c) 2007 Canonical
#
# Written by Jamshed Kakar <jkakar@kakar.ca>
#
# This file is part of AutoPPA.
#
# AutoPPA is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# AutoPPA is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
import tempfile
import os

import pexpect

from autoppa.exception import ChangelogError, EditorError
from autoppa.file import ReleaseSpecificInclude, VersionReplacement
from autoppa.path import match_path_except, match_path_if


class ReleaseBuilder(object):
    """
    A C{ReleaseBuilder} manages the process of creating a release
    branch, making the changes necessary to generate signed source
    content for each Ubuntu-release we need a package for, uploading
    the signed source to Launchpad's personal package archives (PPA)
    service for building, and merging the release branch back into its
    parent branch.
    """

    def __init__(self, name, version, email, branch, repository, ppa, releases,
                 expect=pexpect, system=os.system, logfile=None):
        """
        @param name: The package name.  Note that multiple .deb files
            may be produced for this package.
        @param version: A version string, such as '0.10.12-test3'.
            This will be automatically transformed into a
            release-specific version string, such as
            '0.10.12-dapper1-test3' when the package is built.
        @param email: The name and e-mail address of the package
            change author in the 'John Doe <john@example.com>' format.
        @param branch: The URL of the branch from which packages will
            be made.
        @param repository: The file system path of the repository
            to create the new release branches in.
        @param ppa: The name of the PPA to use, as defined in C{~/.dput.cf}.
        @param releases: A list of Ubuntu releases, such as
            C{['dapper', 'edgy']}.
        @param expect: Optionally, a L{pexpect}-like object.  Defaults
            to L{pexpect}.
        @param system: Optionally, an L{os.system}-like function.
            Defaults to L{os.system}.
        @param logfile: Optionally, a stream to write log output to.
        """
        self.name = name
        self.version = version
        self.email = email
        self.branch = branch
        self.repository = repository
        self.ppa = ppa
        self.releases = releases
        self.working_tree = None
        self.changelog = None
        self.logfile = logfile
        self.expect = expect
        self.system = system
        self.version_replacement = True
        self._prepared_files = set()

    def prepare_changelog(self, filename=None):
        """Prepare the changelog to use for the new packages.

        This method tries to open the user's editor using the VISUAL 
        or EDITOR environment variable.  It loads a changelog template that can
        be used as a starting point when entering notes.  The notes
        collected here are used with each package.

        @param filename: Optionally, the name of the temporary file to
            use when getting changelog details.
        @raises EditorError: Raised when the neither of the environment 
            variables is defined.
        @raises ChangelogError: Raised when the changelog is saved
            without any changes.
        """
        if "VISUAL" in os.environ:
            editor = os.environ["VISUAL"]
        elif "EDITOR" in os.environ:
            editor = os.environ["EDITOR"]
        else:
            raise EditorError("The VISUAL or EDITOR environment variable must "
                              "be set to prepare the changelog.")

        separator = """\
-------------- This line and the following will be ignored --------------\
"""
        changelog_template = """\
  *
%s

      Package: %s
      Version: %s
  Packaged by: %s
Source branch: %s

Please provide notes for your changelog above.  Note that each entry
should start with a '*' indented two-spaces.  Descriptive text for the
entry should be justified, similar to:

  * This is a sample changelog entry that demonstrates the formatting
    style to use when line wrapping occurs.
  * Don't leave any space between entries.

A changelog will be automatically generated for each build with an
Ubuntu release-specific package version string, the notes you specify
above, your e-mail address and the build time.
""" % (separator, self.name, self.version, self.email, self.branch)

        filename = filename or tempfile.mktemp()
        file = open(filename, "w")
        try:
            file.write(changelog_template)
        finally:
            file.close()

        self.system("%s %s" % (editor, filename))

        file = open(filename, "r")
        try:
            changelog = file.read()
        finally:
            file.close()

        changelog = changelog[:changelog.find("\n%s" % separator)]
        end = changelog_template.find("\n%s" % separator)
        if not changelog or changelog == changelog_template[:end]:
            raise ChangelogError("Changelog details must be provided.")
        self.changelog = changelog

    def prepare_working_tree(self):
        """Prepare a working tree for the new release."""
        self.expect.run("mkdir -p %s" % self.repository, logfile=self.logfile)
        self.working_tree = "%s/%s" % (self.repository,
                                       self._get_working_tree())
        self.expect.run("bzr branch %s %s" % (self.branch, self.working_tree),
                        logfile=self.logfile, timeout=None)

    def _get_working_tree(self):
        return "%s_%s" % (self.name, self.version)

    def prepare_custom_files(self, release, datetime_factory=None):
        """Prepare the release-specific files for the new release.

        Packages often require release-specific content for files such
        as the Debian C{control} and C{rules} files.  C{autoppa}
        provides a simple mechanism to deal with this problem.  Any
        file that ends in C{.autoppa} is processed and the output is
        written to the filename minus the C{.autoppa} ending.  The
        processing is simple:

         * All lines are included in the output by default.
         * Any line that begins with C{__AUTOPPA_$release__} is
           included only if C{$release} matches the value of
           C{release}.  Matching is case-sensitive.
         * The magic string C{__AUTOPPA_VERSION__} (found on any
           included line) will be replaced with the release-specific
           package version.

        In addition to special processing of files that end in
        C{.autoppa} an Ubuntu-release specific changelog entry is
        created and added to C{debian/changelog}.  A new changelog
        file is created if necessary.

        @param release: The name of the release to prepare package
            files for.
        @param datetime_factory: Optionally, an object that yields a
            datetime instance to use as the changelog time.  Defaults
            to C{datetime.utcnow}.
        """
        datetime_factory = datetime_factory or datetime.utcnow
        self._prepare_custom_autoppa_files(release)
        if self.version_replacement:
            self._prepare_custom_autoppa_version(release)
        self._prepare_custom_changelog(release, datetime_factory)

    def _prepare_custom_autoppa_files(self, release):
        include = ReleaseSpecificInclude(release)
        for path in match_path_if(self.working_tree, ".*\.autoppa"):
            start = len(self.working_tree)+1
            end = -(len(".autoppa"))
            self._prepared_files.add(path[start:end])
            file = open(path[:-len(".autoppa")], "w")
            try:
                template = open(path, "r")
                try:
                    for i, line in enumerate(template):
                        line = include.process(line, path, i+1)
                        if line is not None:
                            file.write(line)
                finally:
                    template.close()
            finally:
                file.close()

    def _prepare_custom_autoppa_version(self, release):
        version = VersionReplacement(self._get_version(release))
        for path in match_path_except(self.working_tree, ".*\.autoppa"):
            start = len(self.working_tree)+1
            self._prepared_files.add(path[start:])
            file = open(path, "r")
            try:
                lines = [line for line in file]
            finally:
                file.close()

            file = open(path, "w")
            try:
                for i, line in enumerate(lines):
                    line = version.process(line, path, i+1)
                    file.write(line)
            finally:
                file.close()

    def _get_release_number(self, release_name):
        numbers = {"dapper": "6.06", "edgy": "6.10", "feisty": "7.04",
                   "gutsy": "7.10", "hardy": "8.04", "intrepid": "8.10",
                   "jaunty": "9.04", "karmic": "9.10"}
        return numbers[release_name]

    def _get_version(self, release):
        parts = self.version.split("-", 1)
        return "%s-%s.%s" % (parts[0], parts[1],
                             self._get_release_number(release))

    def _prepare_custom_changelog(self, release, datetime_factory):
        try:
            path = os.path.join(self.working_tree, "debian/changelog")
            file = open(path, "r")
            try:
                old_changelog = file.read()
            finally:
                file.close()
        except IOError:
            old_changelog = ""

        version = self._get_version(release)
        date = datetime_factory().strftime("%a, %-d %b %Y %H:%M:%S +0000")
        changelog = """\
%(package)s (%(version)s) %(release)s; urgency=low

%(notes)s

 -- %(packager)s  %(date)s

""" % {"package": self.name,
       "version": version,
       "release": release,
       "notes": self.changelog,
       "packager": self.email,
       "date": date}
        file = open(path, "w")
        try:
            file.write(changelog + old_changelog)
        finally:
            file.close()

    def prepare_signed_source(self):
        """Prepare signed source files that can be uploaded to a PPA."""
        working_directory = os.getcwd()
        os.chdir(self.working_tree)
        try:
            self.expect.run("debuild --no-tgz-check -S", logfile=self.logfile,
                            timeout=None)
        finally:
            os.chdir(working_directory)

    def commit_changes(self, release):
        """Commit changes made to the working tree for C{release}."""
        message = "Updated files for %s_%s." % (self.name,
                                                self._get_version(release))
        working_directory = os.getcwd()
        os.chdir(self.working_tree)
        try:
            self.expect.run('bzr commit -m "%s"' % message,
                            logfile=self.logfile, timeout=None)
        finally:
            os.chdir(working_directory)

    def upload_signed_sources(self, release):
        """Upload signed sources to PPA for C{release}."""
        path = "/".join(self.working_tree.split("/")[:-1])
        filename = "%s_%s_source.changes" % (self.name,
                                             self._get_version(release))

        working_directory = os.getcwd()
        os.chdir(self.repository)
        try:
            self.expect.run("dput %s %s" % (self.ppa,
                                            os.path.join(path, filename)),
                            logfile=self.logfile, timeout=None)
        finally:
            os.chdir(working_directory)

    def merge_changes(self, filename=None):
        """Merge changes from the working tree back into the source branch.

        @param filename: Optionally, the temporary file the merge
            commit message is written to.
        """
        if not filename:
            filename = tempfile.mktemp()
        changelog = """\
Merged %s [a=autoppa].

%s
""" % (self._get_working_tree(), self.changelog)
        file = open(filename, "w")
        try:
            file.write(changelog)
        finally:
            file.close()

        working_directory = os.getcwd()
        os.chdir(self.branch)
        try:
            self.expect.run("bzr merge %s" % self.working_tree,
                            logfile=self.logfile, timeout=None)
            for path in self._prepared_files:
                self.expect.run("bzr revert %s" % path)
            self.expect.run("bzr commit -F %s" % filename,
                            logfile=self.logfile, timeout=None)
        finally:
            os.chdir(working_directory)

    def tag_changes(self):
        """Tag the source branch revision with the merged changes."""
        working_directory = os.getcwd()
        os.chdir(self.branch)
        try:
            self.expect.run("bzr tag %s" % self.version,
                            logfile=self.logfile, timeout=None)
        finally:
            os.chdir(working_directory)

    def cleanup(self):
        """Cleanup signed sources and the release branch."""
        self.expect.run("rm -rf %s" % self.working_tree,
                        logfile=self.logfile, timeout=None)

        for release in self.releases:
            version = self._get_version(release)
            for end in [".dsc", ".tar.gz",
                        "_source.build", "_source.changes", "_source.upload"]:
                filename = "%s_%s%s" % (self.name, version, end)
                path = os.path.join(self.repository, filename)
                self.expect.run("rm '%s'" % path,
                                logfile=self.logfile, timeout=None)


def autoppa_switch(directory, software_version, distro_release):
    """
    Perform AUTOPPA variable replacement on files in C{directory}.
    
    @param directory: Directory in which to search for files to update.
    @param software_version: The new upstream version of the software (see
        C{version} parameter to L{ReleaseBuilder})
    @param distro_release: The Ubuntu release name to use.
    """
    builder = ReleaseBuilder("unused-name", software_version, "unused@email",
                             "unused-branch", "unused-repository", "unused-ppa",
                             [distro_release])
    builder.working_tree = directory
    builder._prepare_custom_autoppa_files(distro_release)
    builder._prepare_custom_autoppa_version(distro_release)
