# Copyright (c) 2007 Canonical
#
# Written by Jamshed Kakar <jkakar@kakar.ca>
#
# This file is part of AutoPPA.
#
# AutoPPA is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# AutoPPA is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from autoppa.exception import FileProcessingError
from autoppa.file import ReleaseSpecificInclude, VersionReplacement


class ReleaseSpecificIncludeTest(unittest.TestCase):

    def setUp(self):
        super(ReleaseSpecificIncludeTest, self).setUp()
        self.processor = ReleaseSpecificInclude("gutsy")

    def test_process_empty_line(self):
        """Empty lines are not modified by the processor."""
        self.assertEquals(self.processor.process(""), "")

    def test_process_line(self):
        """
        Lines without the C{AUTOPPA_INCLUDE():} symbol are included by
        default.
        """
        self.assertEquals(self.processor.process("Hello!"), "Hello!")

    def test_process_matching_release(self):
        """
        A line beginning with C{AUTOPPA_INCLUDE($release):} where
        C{$release} matches the current release is included with the
        symbol removed.
        """
        line = self.processor.process("AUTOPPA_INCLUDE(gutsy):Hello!")
        self.assertEquals(line, "Hello!")

    def test_process_matching_missing_release(self):
        """
        The processor returns C{None} when a line beginning with
        C{AUTOPPA_INCLUDE($releases):} where C{$releases} doesn't
        include a match for the current release is encountered.
        """
        line = self.processor.process("AUTOPPA_INCLUDE(dapper):Hello!")
        self.assertEquals(line, None)

    def test_process_matching_multiple_releases(self):
        """
        More than one release can be included in a
        C{AUTOPPA_INCLUDE():} symbol.  Whitespace is allowed in the
        list of releases.
        """
        line = self.processor.process("AUTOPPA_INCLUDE(edgy,gutsy):Hello!")
        self.assertEquals(line, "Hello!")
        line = self.processor.process("AUTOPPA_INCLUDE( edgy, gutsy ):Hello!")
        self.assertEquals(line, "Hello!")

    def test_process_matching_missing_multiple_releases(self):
        """
        The processor returns C{None} when a line beginning with
        C{AUTOPPA_INCLUDE($releases):} where C{$releases} doesn't
        include a match for the current release is encountered.
        """
        line = self.processor.process("AUTOPPA_INCLUDE(dapper,edgy):Hello!")
        self.assertEquals(line, None)

    def test_process_missing_trailing_parenthesis(self):
        """
        An exception is raised if the trailing parenthesis is not
        included in the C{AUTOPPA_INCLUDE():} symbol.
        """
        self.assertRaises(FileProcessingError, self.processor.process,
                          "AUTOPPA_INCLUDE(dapperHello!")

    def test_process_missing_trailing_colon(self):
        """
        An exception is raised if the trailing colon is not included
        in the C{AUTOPPA_INCLUDE():} symbol.
        """
        self.assertRaises(FileProcessingError, self.processor.process,
                          "AUTOPPA_INCLUDE(dapper)Hello!")


class VersionReplacementTest(unittest.TestCase):

    def setUp(self):
        super(VersionReplacementTest, self).setUp()
        self.processor = VersionReplacement("0.10.11")

    def test_process_missing_symbol(self):
        """
        A line without the AUTOPPA_VERSION(...) symbol is not modified
        by the file processor.
        """
        self.assertEquals(self.processor.process("Hello!"), "Hello!")

    def test_process_empty_version_symbol(self):
        """The AUTOPPA_VERSION() symbol doesn't require a sample version."""
        self.assertEquals(self.processor.process("AUTOPPA_VERSION()"),
                          "AUTOPPA_VERSION(0.10.11)")

    def test_process_sample_version_symbol(self):
        """The AUTOPPA_VERSION() symbol can contain a sample version."""
        self.assertEquals(self.processor.process("AUTOPPA_VERSION(devel)"),
                          "AUTOPPA_VERSION(0.10.11)")

    def test_process_surrounding_parentheses(self):
        """The processor shouldn't modify parentheses surrounding the symbol."""
        self.assertEquals(self.processor.process("(AUTOPPA_VERSION(devel))"),
                          "(AUTOPPA_VERSION(0.10.11))")

    def test_process_missing_trailing_parenthesis(self):
        """
        A L{FileProcessingError} is raised if the trailing parenthesis
        is missing.
        """
        self.assertRaises(FileProcessingError, self.processor.process,
                          "AUTOPPA_VERSION(devel")

    def test_process_python_idiom(self):
        """
        Only the first instance of the C{AUTOPPA_VERSION()} symbol is
        processed to ensure that the following idiom can be used in
        Python code::

            VERSION = 'AUTOPPA_VERSION(devel)'[len('AUTOPPA_VERSION('):-1]
        """
        line = "VERSION = 'AUTOPPA_VERSION(devel)'[len('AUTOPPA_VERSION('):-1]"
        self.assertEquals(
            self.processor.process(line),
            "VERSION = 'AUTOPPA_VERSION(0.10.11)'[len('AUTOPPA_VERSION('):-1]")
