# Copyright (c) 2007 Canonical
#
# Written by Jamshed Kakar <jkakar@kakar.ca>
#
# This file is part of AutoPPA.
#
# AutoPPA is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# AutoPPA is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

from autoppa.path import match_path_except, match_path_if
from autoppa.testing import FilesystemPath


class MatchPathExceptTest(unittest.TestCase):

    def setUp(self):
        super(MatchPathExceptTest, self).setUp()
        self.fs = FilesystemPath()

    def tearDown(self):
        self.fs.tear_down()
        super(MatchPathExceptTest, self).tearDown()

    def test_match_path_except(self):
        """
        L{match_path_except} searches from its starting path and
        yields paths for files that don't match the regular
        expression.
        """
        path = self.fs.make_path(content="demo", path="demo")
        self.fs.make_path(content="test", path="test.autoppa")
        self.assertEquals(list(match_path_except(self.fs.dirname,
                                                 ".*\.autoppa")),
                          [path])

    def test_match_subdirectories(self):
        """L{match_path_except} can find matches in subdirectories."""
        path = self.fs.make_path(content="test", path="test/test")
        self.assertEquals(list(match_path_except(self.fs.dirname,
                                                 ".*\.autoppa")),
                          [path])

    def test_match_ignores_hidden_directories(self):
        """L{match_path_except} ignores hidden directories."""
        self.fs.make_path(content="test", path=".test/test")
        self.fs.make_path(content="test", path=".test/wubble/test")
        self.assertEquals(list(match_path_except(self.fs.dirname,
                                                 ".*\.autoppa")),
                          [])


class MatchPathIfTest(unittest.TestCase):

    def setUp(self):
        super(MatchPathIfTest, self).setUp()
        self.fs = FilesystemPath()

    def tearDown(self):
        self.fs.tear_down()
        super(MatchPathIfTest, self).tearDown()

    def test_match_path_if(self):
        """
        L{match_path_if} searches from its starting path and yields
        paths for files that match the regular expression.
        """
        self.fs.make_path(content="demo", path="demo")
        path = self.fs.make_path(content="test", path="test.autoppa")
        self.assertEquals(list(match_path_if(self.fs.dirname, ".*\.autoppa")),
                          [path])

    def test_match_subdirectories(self):
        """L{match_path_if} can find matches in subdirectories."""
        path = self.fs.make_path(content="test", path="test/test.autoppa")
        self.assertEquals(list(match_path_if(self.fs.dirname, ".*\.autoppa")),
                          [path])

    def test_match_ignores_hidden_directories(self):
        """L{match_path_if} ignores hidden directories."""
        self.fs.make_path(content="test", path=".test/test.autoppa")
        self.fs.make_path(content="test", path=".test/wubble/test.autoppa")
        self.assertEquals(list(match_path_if(self.fs.dirname, ".*\.autoppa")),
                          [])
