# Copyright (c) 2007 Canonical
#
# Written by Jamshed Kakar <jkakar@kakar.ca>
#
# This file is part of AutoPPA.
#
# AutoPPA is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# AutoPPA is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
import os
import unittest

from autoppa.release import (ReleaseBuilder, EditorError, ChangelogError,
                             autoppa_switch)
from autoppa.testing import FakeExpect, FilesystemPath


class ReleaseBuilderTest(unittest.TestCase):

    def setUp(self):
        super(ReleaseBuilderTest, self).setUp()
        self.expect = FakeExpect()
        self.fs = FilesystemPath()
        self.branch_path = self.fs.make_dir("production")
        self.builder = ReleaseBuilder(
            "supertool", "1.3.11-upside1", "John Smith <john@upside.com>",
            self.branch_path, self.fs.dirname, "supertool-ppa",
            ["dapper", "edgy", "feisty"], expect=self.expect,
            system=self.expect.run)

    def tearDown(self):
        self.fs.tear_down()
        super(ReleaseBuilderTest, self).tearDown()

    def test_public_attributes(self):
        """L{ReleaseBuilder} exposes a number of public attributes."""
        self.assertEquals(self.builder.name, "supertool")
        self.assertEquals(self.builder.version, "1.3.11-upside1")
        self.assertEquals(self.builder.email, "John Smith <john@upside.com>")
        self.assertEquals(self.builder.branch, self.branch_path)
        self.assertEquals(self.builder.repository, self.fs.dirname)
        self.assertEquals(self.builder.releases, ["dapper", "edgy", "feisty"])
        self.assertEquals(self.builder.working_tree, None)
        self.assertEquals(self.builder.changelog, None)

    def test_prepare_working_tree(self):
        """
        L{ReleaseBuilder.prepare_working_tree} creates a new release
        working tree in the target repository.  This is where changes
        will be made prior to generating source to upload to PPA.
        """
        self.builder.prepare_working_tree()
        working_tree = "%s/supertool_1.3.11-upside1" % (
                           self.builder.repository)
        self.assertEquals(self.builder.working_tree, working_tree)

        commands = self.expect.commands
        self.assertEquals(len(commands), 2)
        self.assertEquals(commands[0], ("run", "mkdir -p %s" % self.fs.dirname,
                                        {"logfile": None, "timeout": None}))
        command = "bzr branch %s %s" % (self.builder.branch,
                                        self.builder.working_tree)
        self.assertEquals(commands[1], ("run", command,
                                        {"logfile": None, "timeout": None}))

    def test_prepare_changelog_without_editor_environment_variable(self):
        """
        L{ReleaseBuilder.prepare_changelog} relies on a VISUAL or EDITOR
        environment variable.  An L{EditorError} is raised when they're
        not present.
        """
        original_editor = os.environ.get("EDITOR")
        try:
            if original_editor:
                del os.environ["EDITOR"]
            self.assertRaises(EditorError, self.builder.prepare_changelog)
        finally:
            if original_editor is not None:
                os.environ["EDITOR"] = original_editor

    def test_prepare_changelog_without_content(self):
        """
        A L{ChangelogError} is raised if changelog content isn't
        provided by the user.
        """
        original_editor = os.environ.get("EDITOR")
        try:
            if original_editor is None:
                os.environ["EDITOR"] = "emacs"
            self.assertRaises(ChangelogError, self.builder.prepare_changelog)
        finally:
            if original_editor is None:
                del os.environ["EDITOR"]
            else:
                os.environ["EDITOR"] = original_editor

    def test_prepare_changelog(self):
        """
        L{ReleaseBuilder.prepare_changelog} opens the user's editor
        with a template changelog to edit.  The template includes a
        special line to separate changelog content from help text.
        The help text is not included in the result.
        """
        path = self.fs.make_path()
        def run(command, logfile=None, timeout=None):
            file = open(path, "r")
            try:
                changelog_template = file.read()
            finally:
                file.close()

            changelog = """\
  * New upstream release!
%s
""" % "\n".join(changelog_template.splitlines()[1:])

            file = open(path, "w")
            try:
                file.write(changelog)
            finally:
                file.close()

        self.assertEquals(self.builder.changelog, None)
        self.builder.system = run
        original_editor = os.environ.get("EDITOR")
        try:
            if original_editor is None:
                os.environ["EDITOR"] = "emacs"
            self.builder.prepare_changelog(filename=path)
        finally:
            if original_editor is None:
                del os.environ["EDITOR"]
            else:
                os.environ["EDITOR"] = original_editor
        self.assertEquals(self.builder.changelog, "  * New upstream release!")

    def test_prepare_changelog_using_visual_environment_variable(self):
        """
        L{ReleaseBuilder.prepare_changelog} opens the user's editor with a
        template changelog to edit.  If EDITOR isn't defined VISUAL will be
        used.
        """
        path = self.fs.make_path()
        def run(command, logfile=None, timeout=None):
            file = open(path, "r")
            try:
                changelog_template = file.read()
            finally:
                file.close()

            changelog = """\
  * New upstream release!
%s
""" % "\n".join(changelog_template.splitlines()[1:])

            file = open(path, "w")
            try:
                file.write(changelog)
            finally:
                file.close()

        self.assertEquals(self.builder.changelog, None)
        self.builder.system = run
        original_editor = os.environ.get("EDITOR")
        original_visual = os.environ.get("VISUAL")
        try:
            if original_editor is not None:
                del os.environ["EDITOR"]
            if original_visual is None:
                os.environ["VISUAL"] = "emacs"
            self.builder.prepare_changelog(filename=path)
        finally:
            for environ_var, environ_val in [("EDITOR", original_editor),
                                             ("VISUAL", original_visual)]:
                if environ_val is not None:
                    os.environ[environ_var] = environ_val
                elif environ_var in os.environ:
                    del os.environ[environ_var]
        self.assertEquals(self.builder.changelog, "  * New upstream release!")

    def test_prepare_empty_changelog(self):
        """
        If the template text in the changelog is completely removed
        leaving the file empty, an exception should be raised.
        """
        path = self.fs.make_path()
        def run(command, logfile=None, timeout=None):
            open(path, "w").close()

        self.assertEquals(self.builder.changelog, None)
        self.expect.run = run
        original_editor = os.environ.get("EDITOR")
        try:
            if original_editor is None:
                os.environ["EDITOR"] = "emacs"
            self.assertRaises(ChangelogError, self.builder.prepare_changelog,
                              filename=path)
        finally:
            if original_editor is None:
                del os.environ["EDITOR"]
            else:
                os.environ["EDITOR"] = original_editor

    def test_prepare_custom_files(self):
        """
        L{ReleaseBuilder.prepare_custom_files} ensures that the right
        files are in place for the release being built.  It searches
        for files that end in C{.autoppa}, applies release-specific
        include file processing to matching files and writes output to
        files of the same name, but without the C{.autoppa} extension.
        """
        filename = self.fs.make_path(path="debian/control")
        self.fs.make_path(content="AUTOPPA_INCLUDE(dapper):Dapper!\n"
                                  "Generic content\n"
                                  "AUTOPPA_INCLUDE(gutsy):Gutsy!\n",
                          path="debian/control.autoppa")
        self.builder.working_tree = self.fs.dirname
        self.assertFalse(os.path.exists(filename))
        self.builder.prepare_custom_files("dapper")
        file = open(filename, "r")
        try:
            self.assertEquals(file.read(), "Dapper!\nGeneric content\n")
        finally:
            file.close()

    def test_prepare_custom_files_with_version_content(self):
        """
        L{ReleaseBuilder.prepare_custom_files} replaces the magic
        C{AUTOPPA_VERSION()} symbol in lines found in processed files.
        """
        filename = self.fs.make_path(content="version = 'AUTOPPA_VERSION()'",
                                     path="debian/version")
        self.builder.working_tree = self.fs.dirname
        self.builder.prepare_custom_files("dapper")
        file = open(filename, "r")
        try:
            expected = "version = 'AUTOPPA_VERSION(1.3.11-upside1.6.06)'"
            self.assertEquals(file.read(), expected)
        finally:
            file.close()

    def test_prepare_custom_files_with_version_replacement_suppressed(self):
        """
        L{ReleaseBuilder.prepare_custom_files} replaces the magic
        C{AUTOPPA_VERSION()} symbol in lines found in processed files,
        but only if version replacement is enabled.
        """
        filename = self.fs.make_path(content="version = 'AUTOPPA_VERSION()'",
                                     path="debian/version")
        self.builder.version_replacement = False
        self.builder.working_tree = self.fs.dirname
        self.builder.prepare_custom_files("dapper")
        file = open(filename, "r")
        try:
            self.assertEquals(file.read(), "version = 'AUTOPPA_VERSION()'")
        finally:
            file.close()

    def test_prepare_custom_files_with_version_content_skip_autoppa_files(self):
        """
        L{ReleaseBuilder.prepare_custom_files} skips files that end in
        C{.autoppa} when performing version replacement.
        """
        filename = self.fs.make_path(content="version = 'AUTOPPA_VERSION()'",
                                     path="debian/version.autoppa")
        self.builder.working_tree = self.fs.dirname
        self.builder.prepare_custom_files("dapper")
        file = open(filename, "r")
        try:
            self.assertEquals(file.read(), "version = 'AUTOPPA_VERSION()'")
        finally:
            file.close()

    def test_prepare_custom_files_generates_changelog(self):
        """
        In addition to processing files that end in C{.autoppa},
        L{ReleaseBuilder.prepare_custom_files} also generates
        release-specific changes for C{debian/changelog}.  A new file
        will be generated if a changelog doesn't already exist.
        """
        path = self.fs.make_dir(path="debian")
        changelog_filename = os.path.join(path, "changelog")
        self.builder.changelog = "  * Supertool runs 3x faster than before!"
        self.builder.working_tree = self.fs.dirname
        self.assertFalse(os.path.exists(changelog_filename))
        factory = lambda: datetime(2007, 8, 2, 13, 14, 15)
        self.builder.prepare_custom_files("dapper", datetime_factory=factory)
        file = open(changelog_filename, "r")
        try:
            changelog = file.read()
            self.assertEquals(changelog, """\
supertool (1.3.11-upside1.6.06) dapper; urgency=low

  * Supertool runs 3x faster than before!

 -- John Smith <john@upside.com>  Thu, 2 Aug 2007 13:14:15 +0000

""")
        finally:
            file.close()

    def test_prepare_custom_files_updates_changelog(self):
        """
        If a changelog already exists it will be updated by
        L{ReleaseBuilder.prepare_custom_files}.
        """
        path = self.fs.make_dir(path="debian")
        changelog_filename = os.path.join(path, "changelog")
        changelog = """\
supertool (1.3.10-upside1.6.06) dapper; urgency=low

  * Supertool has a new 'super' button that plays a random audio track
    telling you how super-duper you are!

 -- John Smith <john@upside.com>  Wed, 1 Aug 2007 12:13:14 +0000

"""
        file = open(changelog_filename, "w")
        try:
            file.write(changelog)
        finally:
            file.close()

        self.builder.changelog = "  * Supertool runs 3x faster than before!"
        self.builder.working_tree = self.fs.dirname
        factory = lambda: datetime(2007, 8, 2, 13, 14, 15)
        self.builder.prepare_custom_files("dapper", datetime_factory=factory)
        file = open(changelog_filename, "r")
        try:
            changelog = file.read()
            self.assertEquals(changelog, """\
supertool (1.3.11-upside1.6.06) dapper; urgency=low

  * Supertool runs 3x faster than before!

 -- John Smith <john@upside.com>  Thu, 2 Aug 2007 13:14:15 +0000

supertool (1.3.10-upside1.6.06) dapper; urgency=low

  * Supertool has a new 'super' button that plays a random audio track
    telling you how super-duper you are!

 -- John Smith <john@upside.com>  Wed, 1 Aug 2007 12:13:14 +0000

""")
        finally:
            file.close()

    def test_prepare_signed_source(self):
        """
        L{ReleaseBuilder.prepare_signed_source} builds source-only
        packages suitable for upload to PPA.  The current working
        directory is changed during this process and restored when the
        operation is complete.
        """
        self.expect.include_working_directory = True
        self.builder.working_tree = self.fs.dirname
        working_directory = os.getcwd()
        self.builder.prepare_signed_source()
        self.assertEquals(os.getcwd(), working_directory)
        commands = self.expect.commands
        self.assertEquals(len(commands), 1)
        self.assertEquals(commands[0],
                          ("run", "debuild --no-tgz-check -S",
                           {"logfile": None, "timeout": None,
                            "working-directory": self.builder.repository}))
        self.assertEquals(os.getcwd(), working_directory)

    def test_commit_changes(self):
        """
        L{ReleaseBuilder.commit_changes} commits changes in the
        working tree.
        """
        self.expect.include_working_directory = True
        working_directory = os.getcwd()
        self.builder.working_tree = self.fs.dirname
        self.builder.commit_changes("dapper")
        commands = self.expect.commands
        self.assertEquals(len(commands), 1)
        message = "Updated files for supertool_1.3.11-upside1.6.06."
        self.assertEquals(commands[0],
                          ("run", 'bzr commit -m "%s"' % message,
                           {"logfile": None, "timeout": None,
                            "working-directory": self.builder.working_tree}))
        self.assertEquals(os.getcwd(), working_directory)

    def test_upload_signed_sources(self):
        """
        L{ReleaseBuilder.upload_signed_sources} uploads the signed
        sources generates for C{release}.
        """
        self.expect.include_working_directory = True
        working_directory = os.getcwd()
        working_tree = self.fs.make_path(path="supertool_1.3.11-upside1")
        self.builder.working_tree = working_tree
        self.builder.upload_signed_sources("gutsy")
        commands = self.expect.commands
        self.assertEquals(len(commands), 1)
        command = ("dput supertool-ppa "
                   "%s/supertool_1.3.11-upside1.7.10_source.changes"
                   % self.fs.dirname)
        self.assertEquals(commands[0],
                          ("run", command,
                           {"logfile": None, "timeout": None,
                            "working-directory": self.builder.repository}))
        self.assertEquals(os.getcwd(), working_directory)

    def test_merge_changes(self):
        """
        L{ReleaseBuilder.merge_changes} merges the working tree back
        into the source branch.  The current working directory is
        changed during this process and restored when the operation is
        complete.
        """
        self.expect.include_working_directory = True
        path = self.fs.make_path()
        working_tree = self.fs.make_dir(path="supertool_1.3.11-upside1")
        self.builder.working_tree = working_tree
        working_directory = os.getcwd()
        self.builder.merge_changes(filename=path)
        self.assertEquals(os.getcwd(), working_directory)
        commands = self.expect.commands
        self.assertEquals(len(commands), 2)
        self.assertEquals(commands[0],
                          ("run", "bzr merge %s" % working_tree,
                           {"logfile": None, "timeout": None,
                            "working-directory": self.builder.branch}))
        self.assertEquals(commands[1],
                          ("run", "bzr commit -F %s" % path,
                           {"logfile": None, "timeout": None,
                            "working-directory": self.builder.branch}))
        self.assertEquals(os.getcwd(), working_directory)

    def test_merge_changes_generates_useful_commit_message(self):
        """
        L{ReleaseBuilder.merge_changes} should generate a useful
        commit message.
        """
        path = self.fs.make_path()
        self.builder.working_tree = self.fs.dirname
        self.builder.changelog = "  * New upstream release."
        self.builder.merge_changes(filename=path)
        file = open(path, "r")
        try:
            changelog = """\
Merged supertool_1.3.11-upside1 [a=autoppa].

  * New upstream release.
"""
            self.assertEquals(file.read(), changelog)
        finally:
            file.close()

    def test_merge_changes_reverts_dot_autoppa_generated_files(self):
        """
        L{ReleaseBuilder.merge_changes} restores files generated from
        C{.autoppa}-based templates back to their original state
        before the changes are merged.
        """
        changelog_path = self.fs.make_path()
        path = self.fs.make_path(path="debian/control", content="wubble")
        self.fs.make_path(content="AUTOPPA_INCLUDE(dapper):Dapper!\n"
                                  "Generic content\n"
                                  "AUTOPPA_INCLUDE(gutsy):Gutsy!\n",
                          path="debian/control.autoppa")
        self.builder.working_tree = self.fs.dirname
        self.builder.prepare_custom_files("dapper")
        self.builder.merge_changes(filename=changelog_path)
        commands = self.expect.commands
        self.assertEquals(len(commands), 3)
        self.assertEquals(commands[0],
                          ("run", "bzr merge %s" % self.builder.working_tree,
                           {"logfile": None, "timeout": None}))
        expected_path = path[len(self.builder.working_tree)+1:]
        self.assertEquals(commands[1],
                          ("run", "bzr revert %s" % expected_path,
                           {"logfile": None, "timeout": None}))
        self.assertEquals(commands[2],
                          ("run", "bzr commit -F %s" % changelog_path,
                           {"logfile": None, "timeout": None}))

    def test_merge_changes_reverts_version_replaced_files(self):
        """
        L{ReleaseBuilder.merge_changes} restores files with
        C{AUTOPPA_VERSION()} symbols back to their original state
        before the changes are merged.
        """
        changelog_path = self.fs.make_path()
        path = self.fs.make_path(content="version = 'AUTOPPA_VERSION()'",
                                 path="debian/version")
        self.builder.working_tree = self.fs.dirname
        self.builder.prepare_custom_files("dapper")
        self.builder.merge_changes(filename=changelog_path)
        commands = self.expect.commands
        self.assertEquals(len(commands), 3)
        self.assertEquals(commands[0],
                          ("run", "bzr merge %s" % self.builder.working_tree,
                           {"logfile": None, "timeout": None}))
        expected_path = path[len(self.builder.working_tree)+1:]
        self.assertEquals(commands[1],
                          ("run", "bzr revert %s" % expected_path,
                           {"logfile": None, "timeout": None}))
        self.assertEquals(commands[2],
                          ("run", "bzr commit -F %s" % changelog_path,
                           {"logfile": None, "timeout": None}))

    def test_tag_merged_changes(self):
        """
        L{ReleaseBuilder.tag_changes} uses the provided version number
        to tag the revision of the merged changes (in the source
        branch).
        """
        self.expect.include_working_directory = True
        path = self.fs.make_path()
        self.builder.working_tree = self.fs.dirname
        self.builder.changelog = "  * New upstream release."
        working_directory = os.getcwd()
        self.builder.tag_changes()
        self.assertEquals(os.getcwd(), working_directory)
        commands = self.expect.commands
        self.assertEquals(len(commands), 1)
        self.assertEquals(commands[0],
                          ("run", "bzr tag 1.3.11-upside1",
                           {"logfile": None, "timeout": None,
                            "working-directory": self.builder.branch}))
        self.assertEquals(os.getcwd(), working_directory)

    def test_cleanup(self):
        """
        L{ReleaseBuilder.cleanup} removes files generated during the
        build.
        """
        path = self.fs.make_path()
        working_tree = os.path.join(self.fs.dirname, "supertool-1.3.11")
        self.builder.working_tree = working_tree
        self.builder.cleanup()
        commands = self.expect.commands
        self.assertEquals(len(commands), 16)
        self.assertEquals(commands[0],
                          ("run", "rm -rf %s" % working_tree,
                           {"logfile": None, "timeout": None}))

        paths = []
        for begin in ["supertool_1.3.11-upside1.6.06",
                      "supertool_1.3.11-upside1.6.10",
                      "supertool_1.3.11-upside1.7.04"]:
            for end in [".dsc", ".tar.gz",
                        "_source.build", "_source.changes", "_source.upload"]:
                paths.append(begin+end)

        for i, command in enumerate(commands[1:]):
            path = os.path.join(self.builder.repository, paths[i])
            self.assertEquals(command,
                              ("run", "rm '%s'" % path,
                               {"logfile": None, "timeout": None}))


class SwitchTest(unittest.TestCase):
    def setUp(self):
        self.fs = FilesystemPath()

    def tearDown(self):
        self.fs.tear_down()
        super(SwitchTest, self).tearDown()

    def test_autoppa_switch_version(self):
        """
        L{autoppa_switch} replaces AUTOPPA_VERSION usages with the version of
        the software passed.
        """
        path = self.fs.make_path("VERSION = 'AUTOPPA_VERSION(replaceme)'")
        autoppa_switch(self.fs.dirname, "1.0-1", "intrepid")
        self.assertEquals(open(path).read(),
                          "VERSION = 'AUTOPPA_VERSION(1.0-1.8.10)'")

    def test_autoppa_switch_autoppa_includes(self):
        """
        L{autoppa_switch} generates files from .autoppa files, correctly
        choosing values based on C{AUTOPPA_INCLUDE}.
        """
        path = self.fs.make_path("""\
AUTOPPA_INCLUDE(dapper):DAP
AUTOPPA_INCLUDE(intrepid):INT
""", path="foo.autoppa")
        autoppa_switch(self.fs.dirname, "1.0-1", "intrepid")
        generated_foo = os.path.join(self.fs.dirname, "foo")
        self.assertEquals(open(generated_foo).read(), "INT\n")
