# Copyright (c) 2007 Canonical
#
# Written by Jamshed Kakar <jkakar@kakar.ca>
#
# This file is part of AutoPPA.
#
# AutoPPA is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# AutoPPA is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

from autoppa.configuration import Configuration
from autoppa.text import (
    get_options_and_arguments, get_build_target_and_version, create_builders,
    main)
from autoppa.exception import BuildTargetError
from autoppa.testing import FilesystemPath, FakeExpect


class OptionParserTest(unittest.TestCase):

    def test_defaults(self):
        options, arguments = get_options_and_arguments([])
        filename = "%s/.autoppa.conf" % os.environ["HOME"]
        self.assertEquals(options.config, filename)
        self.assertEquals(options.log, None)
        self.assertFalse(options.silent)
        self.assertFalse(options.no_upload)
        self.assertFalse(options.no_merge)
        self.assertFalse(options.no_tag)
        self.assertFalse(options.no_cleanup)

    def test_config(self):
        options, arguments = get_options_and_arguments(["-c", "autoppa.conf"])
        self.assertEquals(options.config, "autoppa.conf")
        options, arguments = get_options_and_arguments(["--config",
                                                        "autoppa.conf"])
        self.assertEquals(options.config, "autoppa.conf")

    def test_log(self):
        options, arguments = get_options_and_arguments(["-l", "autoppa.log"])
        self.assertEquals(options.log, "autoppa.log")
        options, arguments = get_options_and_arguments(["--log", "autoppa.log"])
        self.assertEquals(options.log, "autoppa.log")

    def test_silent(self):
        options, arguments = get_options_and_arguments(["--silent"])
        self.assertEquals(options.silent, True)

    def test_no_upload(self):
        options, arguments = get_options_and_arguments(["--no-upload"])
        self.assertEquals(options.no_upload, True)

    def test_no_merge(self):
        options, arguments = get_options_and_arguments(["--no-merge"])
        self.assertEquals(options.no_merge, True)

    def test_no_tag(self):
        options, arguments = get_options_and_arguments(["--no-tag"])
        self.assertEquals(options.no_tag, True)

    def test_no_cleanup(self):
        options, arguments = get_options_and_arguments(["--no-cleanup"])
        self.assertEquals(options.no_cleanup, True)


class GetBuildTargetAndVersionTest(unittest.TestCase):

    def test_simple_name(self):
        """
        L{get_build_target_and_version} splits a command-line argument
        and returns a 2-tuple of C{(build-target, version)}.
        """
        argument = "supertool-1.1.14-upside1"
        build_target, version = get_build_target_and_version(argument)
        self.assertEquals(build_target, "supertool")
        self.assertEquals(version, "1.1.14-upside1")

    def test_complex_name(self):
        """
        L{get_build_target_and_version} can correctly split package
        names that have '-' characters in their name.
        """
        argument = "fantastic-tool-0.0.1-upside1"
        build_target, version = get_build_target_and_version(argument)
        self.assertEquals(build_target, "fantastic-tool")
        self.assertEquals(version, "0.0.1-upside1")

    def test_malformed_name(self):
        """
        L{get_build_target_and_version} raises L{BuildTargetError} if
        the argument is invalid.
        """
        self.assertRaises(BuildTargetError,
                          get_build_target_and_version, "terrible-tool")


class CreateBuildersTest(unittest.TestCase):

    def setUp(self):
        super(CreateBuildersTest, self).setUp()
        self.fs = FilesystemPath()

    def tearDown(self):
        self.fs.tear_down()
        super(CreateBuildersTest, self).tearDown()

    def test_create_builders(self):
        """
        L{create_builders} returns a list of L{ReleaseBuilder}
        instances suitable for preparing sources for building with PPA
        using a list of build targets and configuration data.
        """
        configuration = """\
[supertool]
email = John Smith <john@upside.com>
branch = /home/john/src/supertool/production
repository = /home/john/src/supertool
ppa = supertool-ppa
releases = dapper edgy feisty gutsy
"""
        path = self.fs.make_path(content=configuration)
        configuration = Configuration.from_file(path)
        arguments = ["supertool-1.3.11-upside1"]
        builders = create_builders(arguments, configuration)
        self.assertEquals(len(builders), 1)
        self.assertEquals(builders[0].name, "supertool")
        self.assertEquals(builders[0].version, "1.3.11-upside1")
        self.assertEquals(builders[0].email, "John Smith <john@upside.com>")
        self.assertEquals(builders[0].branch,
                          "/home/john/src/supertool/production")
        self.assertEquals(builders[0].repository, "/home/john/src/supertool")
        self.assertEquals(builders[0].releases,
                          ["dapper", "edgy", "feisty", "gutsy"])
        self.assertEquals(builders[0].version_replacement, True)

    def test_create_multiple_releases(self):
        """
        L{create_builders} creates multiple L{ReleaseBuilder}
        instances when multiple build targets are specified.  The
        order of releases is maintained.
        """
        configuration = """\
[awesometool]
email = John Smith <john@upside.com>
branch = /home/john/src/awesometool/production
repository = /home/john/src/awesometool
ppa = awesometool-ppa
releases = dapper

[supertool]
email = John Smith <john@upside.com>
branch = /home/john/src/supertool/production
repository = /home/john/src/supertool
ppa = supertool-ppa
releases = dapper edgy feisty gutsy
"""
        path = self.fs.make_path(content=configuration)
        configuration = Configuration.from_file(path)
        arguments = ["supertool-1.3.11-upside1", "awesometool-2.43.2-upside2"]
        builders = create_builders(arguments, configuration)
        self.assertEquals(len(builders), 2)
        self.assertEquals(builders[0].name, "supertool")
        self.assertEquals(builders[0].version, "1.3.11-upside1")
        self.assertEquals(builders[1].name, "awesometool")
        self.assertEquals(builders[1].version, "2.43.2-upside2")

    def test_create_builders_with_unknown_build_target(self):
        configuration = """\
[supertool]
email = John Smith <john@upside.com>
branch = /home/john/src/supertool/production
repository = /home/john/src/supertool
ppa = supertool-ppa
releases = dapper edgy feisty gutsy
"""
        path = self.fs.make_path(content=configuration)
        configuration = Configuration.from_file(path)
        arguments = ["crazytool-0.3.21"]
        self.assertRaises(BuildTargetError, create_builders, arguments,
                          configuration)

    def test_create_release_with_version_replacement_suppressed(self):
        """
        The C{no-version-replacement} option disables version
        replacement.  This is required for AutoPPA to build releases
        of itself.
        """
        configuration = """\
[supertool]
email = John Smith <john@upside.com>
branch = /home/john/src/supertool/production
repository = /home/john/src/supertool
ppa = supertool-ppa
releases = dapper
options = no-version-replacement
"""
        path = self.fs.make_path(content=configuration)
        configuration = Configuration.from_file(path)
        arguments = ["supertool-0.3.21"]
        builders = create_builders(arguments, configuration)
        self.assertEquals(len(builders), 1)
        self.assertEquals(builders[0].version_replacement, False)


class MainTest(unittest.TestCase):

    def setUp(self):
        super(MainTest, self).setUp()
        self.fs = FilesystemPath()

    def test_missing_build_targets(self):
        """
        Build targets are passed to C{autoppa} on the command line, as
        in::

          autoppa landscape-client-0.10.14-landscape20

        If no built targets are provided a L{BuildTargetError} is
        raised.
        """
        self.assertRaises(BuildTargetError, main, arguments=["autoppa"])
