# Copyright (c) 2007 Canonical
#
# Written by Jamshed Kakar <jkakar@kakar.ca>
#
# This file is part of AutoPPA.
#
# AutoPPA is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# AutoPPA is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The interactive text front-end to AutoPPA."""

import sys
import os
from cStringIO import StringIO
from optparse import OptionParser

import pexpect

from autoppa.exception import BuildTargetError
from autoppa.configuration import Configuration
from autoppa.release import ReleaseBuilder
from autoppa import VERSION


def get_options_and_arguments(arguments):
    """Get a 2-tuple of C{(options, arguments)} using C{arguments}."""
    parser = OptionParser(version=VERSION)
    parser.add_option("-c", "--config", metavar="FILE",
                      help="Load configuration details from this file.  "
                           "Default is ~/.autoppa.conf.")
    parser.add_option("-l", "--log", metavar="FILE",
                      help="The file to write the log to.")
    parser.add_option("--no-upload", action="store_true",
                      help="Don't automatically upload signed sources to PPA.")
    parser.add_option("--no-merge", action="store_true",
                      help="Don't automatically merge the changes to the "
                           "source branch.  This implicitly includes --no-tag.")
    parser.add_option("--no-tag", action="store_true",
                      help="Don't automatically tag the merged changes in the "
                           "source branch.")
    parser.add_option("--no-cleanup", action="store_true",
                      help="Don't automatically cleanup signed sources and "
                           "the release branch created during a build.")
    parser.add_option("--silent", action="store_true",
                      help="Suppress internal command output.  This is "
                           "implicit when -l is specified.")
    options, arguments = parser.parse_args(arguments)
    if not options.config:
        options.config = os.path.join(os.environ["HOME"], ".autoppa.conf")
    return options, arguments


def get_build_target_and_version(argument):
    """Get a 2-tuple of C{(build-target, version)} using C{argument}."""
    parts = argument.split("-")
    for i, part in enumerate(parts):
        if part[0].isdigit():
            return "-".join(parts[:i]), "-".join(parts[i:])
    raise BuildTargetError("'%s' is an invalid name." % argument)


def create_builders(arguments, configuration, expect=pexpect, logfile=None):
    """
    Factory method returns a list of L{ReleaseBuilder} instances, one
    per build target specified in the list of arguments passed to
    C{autoppa}.

    @param arguments: A list of arguments provided by the user, such
        as C{['supertool-1.1.14-upside1']}.
    @param configuration: A L{Configuration} instance.
    @param expect: Optionally, a L{pexpect} like object.  Defaults to
        L{pexpect}.
    @param logfile: Optionally, a stream to write log output to.
    @raises BuildTargetError: Raised when a build target is specified
        that doesn't exist in the configuration file.
    """
    builders = []
    for argument in arguments:
        build_target, version = get_build_target_and_version(argument)
        data = configuration.get_build_target(build_target)
        if not data:
            raise BuildTargetError("'%s' is not a known build target."
                                   % build_target)
        builder = ReleaseBuilder(build_target, version, data["email"],
                                 data["branch"], data["repository"],
                                 data["ppa"], data["releases"].split(),
                                 expect=expect, logfile=logfile)
        if "options" in data:
            options = set(data["options"].split())
            if "no-version-replacement" in options:
                builder.version_replacement = False
        builders.append(builder)
    return builders


def main(arguments=sys.argv, expect=pexpect):
    """Get changelog data from the user and upload signed source to PPA."""
    options, option_arguments = get_options_and_arguments(arguments[1:])
    if not option_arguments:
        raise BuildTargetError("No build targets provided.")
    configuration = Configuration.from_file(options.config)
    if options.log:
        logfile = open(options.log, "w")
    elif not options.silent:
        logfile = sys.stdout
    else:
        logfile = StringIO()

    logfile.write("AutoPPA %s\n" % VERSION)

    builders = create_builders(option_arguments, configuration,
                               expect=expect, logfile=logfile)
    for builder in builders:
        logfile.write("\n== Preparing changelog for %s ==\n" % builder.name)
        builder.prepare_changelog()
    for builder in builders:
        logfile.write("\n== Preparing working tree for %s ==\n" % builder.name)
        builder.prepare_working_tree()
        logfile.write("\n== Preparing signed source for %s ==\n" % builder.name)
        for release in builder.releases:
            builder.prepare_custom_files(release)
            builder.prepare_signed_source()
            builder.commit_changes(release)

    if not options.no_upload:
        for builder in builders:
            logfile.write("\n== Uploading signed source for %s ==\n"
                          % builder.name)
            for release in builder.releases:
                builder.upload_signed_sources(release)

    if not options.no_merge:
        for builder in builders:
            logfile.write("\n== Merging changes for %s ==\n" % builder.name)
            builder.merge_changes()
            if not options.no_tag:
                logfile.write("\n== Tagging changes for %s ==\n" % builder.name)
                builder.tag_changes()

    if not options.no_cleanup:
        for builder in builders:
            logfile.write("\n== Cleaning up signed sources and build branch "
                          "for %s ==\n" % builder.name)
            builder.cleanup()

    if options.log:
        logfile.close()

    logfile.write("AutoPPA build finished.\n")
